from typing import Optional

from amino import List, Map, Nothing

from ribosome.machine.transition import may_handle
from ribosome.machine.state import Component, ComponentMachine
from ribosome.unite import UniteSource
from ribosome.unite.data import UniteKind
from ribosome.machine.message_base import Message
from ribosome.machine.machine import Machine
from ribosome.nvim import NvimFacade, NvimIO

from proteome.components.unite.data import UniteSelectAdd, UniteSelectAddAll, UniteProjects, UniteMessage
from proteome.components.unite import UniteNames

addable = UniteSource(UniteNames.addable, UniteNames.addable_candidates, UniteNames.addable, Nothing)
all_addable = UniteSource(UniteNames.all_addable, UniteNames.all_addable_candidates, UniteNames.addable, Nothing)
add_action = Map(name='add', handler=UniteNames.add_project, desc='add project')
add_pro = UniteKind(UniteNames.addable, List(add_action))
projects = UniteSource(UniteNames.projects, UniteNames.projects_candidates, UniteNames.project, Nothing)
delete_action = Map(name='delete', handler=UniteNames.delete_project, desc='delete project')
activate_action = Map(name='activate', handler=UniteNames.activate_project, desc='activate project', is_selectable=0)
project = UniteKind(UniteNames.project, List(activate_action, delete_action))
entities = List(addable, all_addable, add_pro, projects, project)


class UniteTransitions(Component):

    def unite_cmd(self, cmd: str) -> None:
        args = ' '.join(self.msg.unite_args)
        self.vim.cmd('Unite {} {}'.format(cmd, args))

    @may_handle(UniteSelectAdd)
    def select_add(self) -> None:
        self.unite_cmd(UniteNames.addable)

    @may_handle(UniteSelectAddAll)
    def select_add_all(self) -> None:
        self.unite_cmd(UniteNames.all_addable)

    @may_handle(UniteProjects)
    def projects(self) -> None:
        self.unite_cmd(UniteNames.projects)


class Plugin(ComponentMachine):

    def __init__(self, vim: NvimFacade, name: Optional[str], parent: Optional[Machine]) -> None:
        super().__init__(vim, UniteTransitions, name, parent)
        self._unite_ready = False

    def prepare(self, msg: Message) -> None:
        if not self._unite_ready and isinstance(msg, UniteMessage):
            self._setup_unite()

    def _setup_unite(self) -> None:
        entities.traverse(lambda a: NvimIO(a.define), NvimIO).attempt(self.vim)
        self._unite_ready = True

__all__ = ('Plugin')



from ribosome.machine.state import RootMachine
from ribosome.nvim import HasNvim
from ribosome.machine.modular import ModularMachine
from ribosome.machine.transitions import Transitions
from ribosome.machine.base import MachineBase
from ribosome.nvim import NvimFacade

from proteome.logging import Logging


class ProteomeComponent(ModularMachine, HasNvim, Logging):

    def __init__(self, vim: NvimFacade, parent=None, title=None) -> None:
        MachineBase.__init__(self, parent, title=title)
        HasNvim.__init__(self, vim)


class ProteomeState(RootMachine, Logging):

    @property
    def title(self):
        return 'proteome'


class ProteomeTransitions(Transitions, HasNvim):

    def __init__(self, machine, *a, **kw):
        Transitions.__init__(self, machine, *a, **kw)
        HasNvim.__init__(self, machine.vim)

__all__ = ('ProteomeComponent', 'ProteomeState', 'ProteomeTransitions')



from ribosome.machine.message_base import message

Gen = message('Gen', 'project')
GenAll = message('GenAll')
Kill = message('Kill')
CurrentBuffer = message('CurrentBuffer')

__all__ = ('Gen', 'GenAll', 'Kill', 'CurrentBuffer')



from amino import List, Map, __, _, Path, Nil, Either, Lists, L, Try, Right, do
from amino.boolean import true

from ribosome.settings import (PluginSettings, path_setting, path_list_setting, setting_ctor, path_list, str_setting,
                               bool_setting)


config_path_help = '''Each json file in this directory is read to populate the list of project configurations.
Here you can either define independent projects that can be added with `ProAdd!`:
```
{
"name": ".sbt",
"type": "scala",
"langs": ["scala"],
"root": "~/.sbt/0.13"
}
```
Or you can amend projects that are located in one of the project base dirs, for example to set additional languages.
'''

base_dirs_help = '''A list of directories that are searched for projects of the `<type>/<name>` structure. Types are
categories grouping projects by language or other, arbitrary, criteria. When adding a project with `ProAdd! type/name`,
it is matched against these paths.
'''

type_base_dirs_help = '''A dictionary of directories mapped to lists of strings defining project types.
A directory is searched when adding projects of a type matching one of the corresponding types.
'''

tags_command_help = '''If your project uses an alternative program to generate its tags, set this variable to its
executable path or name. Arguments go into `g:proteome_tags_args`.
'''

tags_args_help = '''Set this to a space-separated list of arguments for the custom tags command.
You can supply a python format string containing the variables `langs`, `tag_file` and `root`.
'''

load_buffers_help = '''On invocation of `ProLoad`, proteome persists the current buffer list to disk. If this flag is
true, they will be restored upon startup.
'''


@do
def cons_type_base_dirs(data: dict) -> Either[str, Map[Path, List[str]]]:
    keys, values = Lists.wrap(data.items()).unzip
    paths = yield path_list(keys)
    types = yield values.traverse(__.traverse(L(Try)(Path, _), Either), Either)
    yield Right(Map(paths.zip(types)))


type_base_dirs_setting = setting_ctor(dict, cons_type_base_dirs)


class ProteomeSettings(PluginSettings):

    def __init__(self) -> None:
        super().__init__('proteome')
        self.config_path = path_setting('config_path', 'config directory', config_path_help, True,
                                        Right(Path('/dev/null')))
        self.base_dirs = path_list_setting('base_dirs', 'project base dirs', base_dirs_help, True, Right(Nil))
        self.type_base_dirs = type_base_dirs_setting('type_base_dirs', 'project type base dir map', type_base_dirs_help,
                                                     True, Right(Nil))
        self.tags_command = str_setting('tags_command', 'custom command for ctags generation', tags_command_help, True)
        self.tags_args = str_setting('tags_args', 'args for custom ctags command', tags_args_help, True)
        self.load_buffers = bool_setting('load_buffers', 'load persisted buffers on startup', load_buffers_help, True,
                                         Right(true))

__all__ = ('ProteomeSettings',)



from proteome.components.core.message import (Add, RemoveByIdent, Create, Next, Prev, SetProject, SetProjectIdent,
                                           SetProjectIndex, SwitchRoot, Save, Added, Removed, ProjectChanged, BufEnter,
                                           Initialized, MainAdded, Show, AddByParams, CloneRepo)

__all__ = ('Add', 'RemoveByIdent', 'Create', 'Next', 'Prev', 'SetProject', 'SetProjectIdent', 'SetProjectIndex',
           'SwitchRoot', 'Save', 'Added', 'Removed', 'ProjectChanged', 'BufEnter', 'Initialized', 'MainAdded', 'Show',
           'AddByParams', 'CloneRepo')



from ribosome.machine.message_base import Message


class UniteMessage(Message, varargs='unite_args'):
    pass


class UniteSelectAdd(UniteMessage):
    pass


class UniteSelectAddAll(UniteMessage):
    pass


class UniteProjects(UniteMessage):
    pass


class UniteNames():
    addable_candidates = '_proteome_unite_addable'
    all_addable_candidates = '_proteome_unite_all_addable'
    projects_candidates = '_proteome_unite_projects'

    add_project = '_proteome_unite_add_project'
    delete_project = '_proteome_unite_delete_project'
    activate_project = '_proteome_unite_activate_project'

    addable = 'proteome_addable'
    all_addable = 'proteome_all_addable'
    projects = 'proteome_projects'
    project = 'proteome_project'

__all__ = ('UniteNames', 'UniteSelectAdd', 'UniteSelectAddAll', 'UniteProjects')



from proteome.components.unite.data import UniteSelectAdd, UniteSelectAddAll, UniteProjects, UniteNames
from proteome.components.unite.main import Plugin

__all__ = ('Plugin', 'UniteSelectAdd', 'UniteSelectAddAll', 'UniteProjects', 'UniteNames')
